/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: out.c,v 1.28 2021/03/14 18:26:47 babolo Exp $"

#define BLIN_COMPAT     3
#define MIFE_COMPAT     5
#define Bpars_COMPAT    4
#define RECOBE_COMPAT   4

#include <sys/types.h>
#include <sys/event.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <zlib.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "pngtr.h"

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                        ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_oout(pngtr_conf *cf, u_int chctr, u_int opctr, BLIN_flag f, int fd) {                      /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    const pngtr_chidef *opdef;
    u_int               opnum;
    pngtr_operand      *op     = &chunk->operands[opctr];
    int                 ex    = EX_OK;

    ifBLIN_QX3("+ %08X %08X q%u c%u", cf->flags, f, cf->num, chctr);
    chdef = &pngtr_chdescr[chunk->type];
    opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
    opnum = opdef->opn;
    if  (!op->val) goto out;
    if  ((PNGTR_OPERSEL & convey->flags) && !(PNGTR_SELECTD & op->flags)) goto out;
    if  (  (1 < opctr)
        && (  ((PNGTR_CHDEP & opdef->op[1].var) && (op->val[0] > 8))
           || ((PNGTR_CHZIP & opdef->op[1].var) && !!op->val[0])
        )  ) {                                /*  sPLT0  iTXt     */
        opdef = chdef->ch[1];
    }
    if  (PNGTR_PRINTXD & f) {
        switch(opdef->op[opctr].otype) {
        case PNGTR_DU:
        case PNGTR_DD4:
        case PNGTR_DF4:
        case PNGTR_DF2:
        case PNGTR_DW2:
        case PNGTR_DW1:
        case PNGTR_DCHAR:
        case PNGTR_DTFIX:
        case PNGTR_DW1S:
        case PNGTR_DW1R:
        case PNGTR_DU1W:
        case PNGTR_DU1C:
        case PNGTR_DUS:
        case PNGTR_DUC:
        case PNGTR_DUP:
        case PNGTR_DU1N:
        case PNGTR_DW1T:
        case PNGTR_DU1I:
        case PNGTR_DC1:
        case PNGTR_DC2:
            PNGTR_W( (op->size != (ssize_t)(ex = mife_writ(fd, op->val, op->size)))
                   , ("mife_writ %d <> %u @ %s q%u c%u o%u"
                     , ex
                     , op->size
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                     )
                   , EX_IOERR
                   );
            ex = EX_OK;
            break;
        case PNGTR_DCHAZ:
        case PNGTR_DBHZ:
        case PNGTR_DBZ:
        case PNGTR_DBYTE:
        case PNGTR_TRASH:
        case PNGTR_DCC:
        case PNGTR_DCA:
        case PNGTR_DUH:
        case PNGTR_DCA1:
        case PNGTR_DCA2:
            ex = mife_writ(fd, op->val, op->size * op->maxindex);
            PNGTR_W( (op->size * op->maxindex != (ssize_t)ex)
                   , ("mife_writ %d <> %u @ %s q%u c%u o%u"
                     , ex
                     , op->size * op->maxindex
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                     )
                   , EX_IOERR
                   );
            ex = EX_OK;
            break;
    }   }
    if  (PNGTR_PRINTEX & f) {
        switch(opdef->op[opctr].otype) {
        case PNGTR_DBZ:
            if  ((!chunk->infla) && !!(ex = pngtr_infla(cf, chctr, opctr))) {
                ifBLIN_QW0( "infla %s q%u c%u o%u"
                          , !convey->png.nm ? "" : convey->png.nm
                          , cf->num
                          , chctr
                          , opctr
                          );
            }
        /* FALLTHRU */
        case PNGTR_DBHZ:
            ex = mife_writ(fd, chunk->infla, chunk->inflen);
            PNGTR_W( ((0 > ex) || (chunk->inflen != (size_t)ex))
                   , ( "mife_writ %d <> %"BLIN_U" @ %s q%u c%u o%u"
                     , ex
                     , chunk->inflen
                     , !convey->png.nm ? "" : convey->png.nm
                     , cf->num
                     , chctr
                     , opctr
                     )
                   , EX_IOERR
                   );
            ex = EX_OK;
            break;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****                                                                     ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_vchu(pngtr_conf *cf, pngtr_chunk *chunk) {                                                 /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_operand      *op;
    int                 ex    = EX_OK;
    u_char             *p;

    ifBLIN_QX3("+ %08X %08X q%u", cf->flags, chunk->flags, cf->num);
    free(chunk->vchu);
    chunk->vchu = NULL;
    chunk->csz = 0;
    chunk->offset = 0;
    if  (!!chunk->operands) {
        for (u_int opctr = 0; opctr <= chunk->carg; ++opctr) {
            op = &chunk->operands[opctr];
            if  ((PNGTR_OPERSEL & convey->flags) && !(PNGTR_SELECTD & op->flags)) continue;
            if  (!!op->val) chunk->csz += op->size * op->maxindex;
    }   }
    chunk->csz += 12;
    PNGTR_W(!(chunk->vchu = malloc(chunk->csz)),                  ("malloc %u", chunk->csz), EX_OSERR);
    p = (u_char *)chunk->vchu;
    recobe_store(p, -4, chunk->csz - 12);
    p += 4;
    bcopy(pngtr_chdescr[chunk->type].nm, p, 4);
    p += 4;
    if  (!!chunk->operands) {
        for (u_int opctr = 0; opctr <= chunk->carg; ++opctr) {
            op = &chunk->operands[opctr];
            if  ((PNGTR_OPERSEL & convey->flags) && !(PNGTR_SELECTD & op->flags)) continue;
            if  (!!op->val) bcopy(op->val, p, op->size * op->maxindex);
            p += op->size * op->maxindex;
    }   }
    recobe_store(CRC(chunk->vchu), -4, recobe_crc32((u_char *)chunk->vchu->nm, chunk->csz - 8));
    recobe_store(p, -4, recobe_crc32((u_char *)chunk->vchu->nm, chunk->csz - 8));
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                           ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_cout(pngtr_conf *cf, u_int chctr, BLIN_flag f, int fd) {                                   /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey       *convey = &cf->convey[cf->num];
    pngtr_chunk        *chunk = &convey->chunks[chctr];
    const pngtr_chdef  *chdef;
    const pngtr_chidef *opdef;
    u_int               opnum;
    pngtr_hdr          *hdr;
    int                 ex    = EX_OK;
    u_int8_t           *c = convey->png.cont;

    ifBLIN_QX3("+ %08X %08X q%u c%u", cf->flags, f, cf->num, chctr);
    chdef = &pngtr_chdescr[chunk->type];
    opdef = chdef->ch[(PNGTR_CHCOLOR & chdef->shift) ? convey->IHDR.colortype : 0];
    opnum = opdef->opn;
    if  (!!chunk->operands) for (u_int opctr = 0; opctr < opnum; ++opctr) {
    ;   pngtr_operand *op = &chunk->operands[opctr];

    ;   if  (!op->val) continue;
                                                   /*  pCAL     */
    ;   if  (PNGTR_OPNUM & opdef->op[opctr].var) opnum = 6 + op->val[0];
    ;   PNGTR_O(!!(ex = pngtr_oout(cf, chctr, opctr, f, fd)),                                ("oout"));
    }
    hdr = (pngtr_hdr *)&c[chunk->offset];
    if  (PNGTR_PRINTCD & f) {
        if  (PNGTR_VIRTUAL & chunk->flags) {
            PNGTR_O(!!(ex = pngtr_vchu(cf, chunk)),                                          ("vchu"));
            hdr = (pngtr_hdr *)chunk->vchu;
        }
        if  (!!chunk->csz) {
            ex = mife_writ(fd, hdr, chunk->csz);
            PNGTR_W((chunk->csz != (ssize_t)ex),     ("mife_writ %d <> %u", ex, chunk->csz), EX_IOERR);
            ex = EX_OK;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    fd  cf->num    f                                           ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_out(pngtr_conf *cf, BLIN_flag f, int fd) {                                                 /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    int               ex    = EX_OK;
    u_int8_t         *c;

    ifBLIN_QX3("+ %08X %08X %u", cf->flags, f, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    convey = &cf->convey[cf->num];
    c = convey->png.cont;
    ifBLIN_QX4("%"BLIN_X" %u %02X", BLIN_I(c), convey->chunum, convey->IHDR.colortype);
    for (u_int chctr = 0; chctr <= convey->chunum; ++chctr) {
        pngtr_chunk        *chunk;

    ;   chunk = &convey->chunks[chctr];
    ;   if  (!(  ((PNGTR_CHAINSL & convey->flags) && (PNGTR_SELECTD & chunk->flags))
              || ((PNGTR_CHATYPE & convey->flags) && ((1 << chunk->type) & convey->typemask))
              || !((PNGTR_CHAINSL | PNGTR_CHATYPE) & convey->flags)
            ) ) {
    ;       continue;
    ;   }
    ;   PNGTR_O(!!(ex = pngtr_cout(cf, chctr, f, fd)),                                       ("cout"));
    }
    if  (PNGTR_PRINTFM & f) {
        PNGTR_W((convey->nowsz != mife_writ(fd, convey->imda, convey->nowsz)),     ("writ"), EX_IOERR);
    }
    if  (PNGTR_PRINTIM & f) {
        if  (!convey->img.cont) {
            PNGTR_O( !!(ex = pngtr_flimg(cf, NULL))
                   , ("pngtr_flimg %s", !convey->png.nm ? "" : convey->png.nm)
                   );
        }
        PNGTR_W( (convey->img.size != mife_writ(fd, convey->img.cont, convey->img.size))
               , ("mife_writ %s", !convey->png.nm ? "" : convey->png.nm)
               , EX_IOERR
               );
    }
    if  (PNGTR_OUTPAM & f) {
        if  (!convey->img.cont) {
            PNGTR_O( !!(ex = pngtr_flimg(cf, NULL))
                   , ("pngtr_flimg %s", !convey->png.nm ? "" : convey->png.nm)
                   );
        }
        pngtr_outpam(cf, fd);
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     cf->ounm  cf->num                                      ****
 *****************************************************************************************************
                                                                                                  ****/
pngtr_dir(pngtr_conf *cf) {                                                                      /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    pngtr_convey     *convey;
    int               ex    = EX_OK;
    u_int8_t         *c;

    ifBLIN_QX3("+ %08X %08X %u", cf->flags, cf->num);
    PNGTR_X((cf->num >= PNGTR_CONVEYNUM),            ("Ill convey num %d", cf->num), ERANGE, EX_USAGE);
    PNGTR_X(!cf->ounm,                                             ("No dir name"), EDOOFUS, EX_USAGE);
    if  ((0 > mkdir(cf->ounm, 0777)) && (EEXIST != errno)) {
        ifBLIN_QW0("mkdir %s", cf->ounm);
        ex = EX_IOERR;
        goto out;
    }
    convey = &cf->convey[cf->num];
    c = convey->png.cont;
    ifBLIN_QX4("%"BLIN_X" %u %02X", BLIN_I(c), convey->chunum, convey->IHDR.colortype);
    for (u_int chctr = 0; chctr <= convey->chunum; ++chctr) {
        pngtr_chunk      *chunk;
        pngtr_hdr        *hdr;
        char             *fn;
        int               fd;
        ssize_t           e;

    ;   chunk = &convey->chunks[chctr];
    ;   hdr = (pngtr_hdr *)&c[chunk->offset];
    ;   ifBLIN_QX4("%u %"BLIN_X" %u", chctr, BLIN_I(hdr), chunk->csz);
    ;   if  (!(  ((PNGTR_CHAINSL & convey->flags) && (PNGTR_SELECTD & chunk->flags))
              || ((PNGTR_CHATYPE & convey->flags) && ((1 << chunk->type) & convey->typemask))
              || !((PNGTR_CHAINSL | PNGTR_CHATYPE) & convey->flags)
            ) ) {
    ;       continue;
    ;   }
    ;   PNGTR_W((0 > (asprintf(&fn, "%s/%05u", cf->ounm, chctr))),             ("asprintf"), EX_IOERR);
    ;   PNGTR_W( (0 > (fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC, 0666)))
               , ("open %05u", chctr)
               , EX_IOERR
               );
    ;   PNGTR_W( ((PNGTR_ADDSIGN & cf->flags) && (8 != mife_writ(fd, pngtr_sg, 8)))
               , ("mife_writ");
               , EX_IOERR
               );
    ;   if  (!!chunk->csz) {
    ;       PNGTR_W( (chunk->csz != (e = mife_writ(fd, hdr, chunk->csz)))
                   , ("mife_writ %d <> %d", (int)e, chunk->csz)
                   , EX_IOERR
                   );
    ;   }
    ;   close(fd);
    ;   free(fn);
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
